@@ -20,6 +20,8 @@
 #define WIN98_EXTENDED_PARTITION 0x0f
 
 #define LINUX_SWAP_PARTITION		0x82
+#define LINUX_EXT2_PARTITION		0x83
+#define PALO_PARTITION			0xf0
 #define LINUX_RAID_PARTITION		0xfd	/* autodetect RAID partition */
 #define LINUX_OLD_RAID_PARTITION	0x86
 
