@@ -17,6 +17,8 @@
 #include <string.h>
 #include <ctype.h>
 #include <sys/mman.h>
+#include <sys/types.h>
+#include <sys/wait.h>
 #if LONG_OPTIONS
 # define _GNU_SOURCE
 # include <getopt.h>
@@ -30,6 +32,7 @@
 static const char Id[] = "$Id: palo.c,v 1.12.4.1 2001/06/13 03:13:17 bame Exp $";
 
 static int Install = 0;
+int verbose = 0;
 
 /* compute the sum of words in an 4-byte aligned region */
 int
@@ -39,7 +42,7 @@ checksum(void *p, size_t len)
     int *x = (int *)p;
     int i;
 
-    if (0) printf("checksum(%p, %u) = ", p, len);
+    if (verbose) printf("checksum(%p, %u) = ", p, len);
     len /= 4;
 
     for (i = 0; i < len; i++)
@@ -47,7 +50,7 @@ checksum(void *p, size_t len)
 	xsum += B32(x[i]);
     }
 
-    if (0) printf("0x%08x\n", xsum);
+    if (verbose) printf("0x%08x\n", xsum);
 
     return (xsum);
 }
@@ -74,7 +77,7 @@ check_bootloader(int media, int line)
     int xsum;
     unsigned int *ipl;
 
-    if (0) printf("check_bootloader %d\n", line);
+    if (verbose) printf("check_bootloader %d\n", line);
 
     STRUCTREAD(media, f, 0);
 
@@ -173,9 +176,11 @@ write_bootloader(int media, int bootloader,
 	error(12);
 
     /* make sure max size boot loader would fit */
-    if (where + MAXBLSIZE >= f0end)
+    if (where + MAXBLSIZE > f0end) {
+	printf("where %d, where+MAX=%d f0end=%d\n",
+	       where, where + MAXBLSIZE, f0end);
 	error(9);
-
+    }
     /* load the boot loader into RAM */
     rblsize = fsize(bootloader);
     assert((rblsize % FW_BLOCKSIZE) == 0);
@@ -425,6 +430,110 @@ do_cdrom(int media, int kernel32, int kernel64,
 
 #define KSIZE (2 * 1024 * 1024)
 
+/* Blocksize for the ext2/3 fs in the palo partition */
+#define EXT2_BLOCKSIZE	1024
+
+/* size in ext2 blocks of hole for bootloader */
+#define EXT2_HOLE	((MAXBLSIZE + 1) / EXT2_BLOCKSIZE)
+
+/* offset in bytes before start of hole,  ext2 doesn't allow holes at
+ * to cover the first four blocks of the filesystem */
+#define EXT2_OFFSET	(4*EXT2_BLOCKSIZE)
+
+void
+do_formatted(int init, int media, const char *medianame, int partition,
+	     int f0start, int f0length, int bootloaderfd, int do_format,
+	     const char *commandline)
+{
+    char partitionname[256];
+    struct firstblock f;
+    int partitionfd;
+
+    /* FIXME: assuming we can simply add the partition to the end
+     * of the whole disc device isn't always true */
+    snprintf(partitionname, sizeof(partitionname), "%s%d", medianame,
+	     partition);
+
+    if ((partitionfd = open(partitionname, O_RDWR)) < 0) {
+	perror(partitionname);
+	exit(1);
+    }
+    close(partitionfd);
+
+    if (init) {
+	/* make the bootloader align at 256k */
+	unsigned int holestart = (f0start + 0x3ffff + EXT2_OFFSET) & ~0x3ffff;
+	unsigned int partition_offset = holestart - f0start;
+	char badblockfilename[256];
+	int fd, i;
+	char cmd[512];
+
+	if (verbose)
+	    printf("f0 partition starts %d, hole %d-%d, end %d\n",
+		   f0start, holestart, holestart + EXT2_HOLE*EXT2_BLOCKSIZE,
+		   f0start + f0length);
+
+	if(partition_offset + EXT2_HOLE > f0length)
+	    error(14, "bootloader");
+
+	sprintf(badblockfilename, "/tmp/paloblk-%d", getpid());
+	if ((fd = open(badblockfilename, O_RDWR | O_CREAT | O_TRUNC)) < 0) {
+	    perror(badblockfilename);
+	    exit(1);
+	}
+    
+	for (i = (holestart - f0start)/EXT2_BLOCKSIZE;
+	     i < (holestart - f0start)/EXT2_BLOCKSIZE + EXT2_HOLE; i++) {
+	    char buf[128];
+	    sprintf(buf, "%d\n", i);
+	    write(fd, buf, strlen(buf));
+	}
+
+	sprintf(cmd, "mke2fs %s -b %d -l %s %s", do_format == 3 ? "-j" : "",
+		EXT2_BLOCKSIZE, badblockfilename, partitionname);
+
+	if (verbose)
+	    printf("Executing: %s\n", cmd);
+	else
+	    strcat(cmd, " > /dev/null 2>&1");
+
+	i = system(cmd);
+	unlink(badblockfilename);
+
+	if(WEXITSTATUS(i) != 0)
+	    error(19, WEXITSTATUS(i));
+
+	
+
+	STRUCTREAD(media, f, 0);
+	fb_init(&f);
+	f.flags |= PFLAG_EXT2;
+
+	if(commandline)
+	    strncpy(f.cmdline, commandline, sizeof(f.cmdline));
+
+	write_bootloader(media, bootloaderfd, holestart,
+			 holestart + EXT2_HOLE*EXT2_BLOCKSIZE, &f);
+	
+	STRUCTWRITE(media, f, 0);
+
+    } else {
+	STRUCTREAD(media, f, 0);
+
+	if(verbose)
+	    printf("Updating formatted ver=%d, start=%d\n",
+		   f.version, f.ipl_addr);
+
+	if(commandline)
+	    strncpy(f.cmdline, commandline, sizeof(f.cmdline));
+
+	write_bootloader(media, bootloaderfd, f.ipl_addr,
+			 f.ipl_addr + EXT2_HOLE*EXT2_BLOCKSIZE, &f);
+
+	STRUCTWRITE(media, f, 0);
+    }
+}
+
 void
 do_randomaccess(int init, int media, int kernel32, int kernel64,
 	const char *commandline, int bootloader, int ramdisk,
@@ -433,7 +542,7 @@ do_randomaccess(int init, int media, int kernel32, int kernel64,
     struct firstblock f;
     int bstart = 0;
 
-    if (0) printf("do_ra(%d, %d, %d, %d, '%s', %d, %d, %u, %u)\n",
+    if (verbose) printf("do_ra(%d, %d, %d, %d, '%s', %d, %d, %u, %u)\n",
 	init, media, kernel32, kernel64, commandline, bootloader, ramdisk, f0start,
 	f0length);
 
@@ -552,9 +661,11 @@ static struct option Longopts[] =
     {"bootloader", 1, 0, 'b'},
     {"ramdisk", 1, 0, 'r'},
     {"init-partitioned", 1, 0, 'I'},
+    {"format-as", 1, 0, 'e'},
     {"update-partitioned", 1, 0, 'U'},
     {"init-tape", 1, 0, 's'},
     {"init-cdrom", 1, 0, 'C'},
+    {"verbose", 0, 0, 'v'},
     {"help", 0, 0, '?'},
     {0, 0, 0, 0}
 };
@@ -583,10 +694,10 @@ main(int argc, char *argv[])
     char *bootloaderfile = "/usr/share/palo/iplboot";
     int init = 0;
     extern char *bld_info;
-    const char gargs[] = "f:C:s:b:k:c:r:I:?";
+    const char gargs[] = "f:C:s:b:k:c:r:I:e:U:v?";
     char *config_file = "/etc/palo.conf";
     char *newargv[MAXARGS];
-    int newargc;
+    int newargc, format_as = 0;
     FILE *fconfig;
 
     assert(sizeof (struct firstblock) == 2048);
@@ -717,9 +828,20 @@ main(int argc, char *argv[])
 	    break;
 	case 'c':
 	    commandline = optarg;
-	    if (strlen(commandline) > 255)
+	    if (strlen(commandline) > 127)
 		error(3);
 	    break;
+	case 'e':
+	    if(strcmp(optarg, "2") == 0)
+		format_as = 2;
+	    else if(strcmp(optarg, "3") == 0)
+		format_as = 3;
+	    else
+		error(0, argv[0]);
+	    break;
+	case 'v':
+	    verbose = 1;
+	    break;
 	default:
 	    error(0, argv[0]);
 	    break;
@@ -758,6 +880,7 @@ main(int argc, char *argv[])
 	    struct diskpartition ptab[MAXPARTS];
 	    int partitioned;
 	    int f0;
+	    struct firstblock f;
 
 	    memset(ptab, 0, sizeof ptab);
 
@@ -782,18 +905,39 @@ main(int argc, char *argv[])
 	    /* find the F0 partition */
 	    for (f0 = 0; f0 < MAXPARTS; f0++)
 	    {
-		if (ptab[f0].id == 0xf0)
+		if (ptab[f0].id == PALO_PARTITION)
 		    break;
 	    }
 	    if (f0 == MAXPARTS)
 		error(11);
-	    print_ptab_pretty(ptab, MAXPARTS);
+	    if(verbose) {
+		print_ptab_pretty(ptab, MAXPARTS);
 
-	    if (0) printf("F0 partition start sector %d length %d\n",
+		printf("F0 partition start sector %d length %d\n",
 		    ptab[f0].start, ptab[f0].length);
+	    }
+	    if (!init) {
+		    STRUCTREAD(media, f, 0);
+		    if (f.version >= 4 && (f.flags & PFLAG_EXT2))
+			/* need to update formatted partition */
+			format_as = 1;
+	    }
 
-	    do_randomaccess(init, media, kernel32, kernel64, commandline, bootloader,
-		ramdisk, ptab[f0].start * 512, ptab[f0].length * 512);
+	    if (format_as) {
+		/* if we're going to be a formatted partition, we can't
+		 * load anything into it, so check we haven't been asked
+		 * to */
+		if(kernel32 != -1 || kernel64 != -1
+		   || ramdisk != -1)
+		    error(18);
+		printf("OK we're doing a format as ext%d\n", format_as);
+		do_formatted(init, media, medianame, f0 + 1,
+			     ptab[f0].start * 512, ptab[f0].length * 512,
+			     bootloader, format_as, commandline);
+	    } else
+		do_randomaccess(init, media, kernel32, kernel64, commandline,
+				bootloader, ramdisk, ptab[f0].start * 512,
+				ptab[f0].length * 512);
 	}
 	break;
     case CDROM:
